/***************************************************
 * 
 * X11 のウィンドウのスクリーンショットを PNG に出力
 *
 * 使用・改変などは自由です。
 * 使用によって生じた損害に関しては一切責任を負いません。
 * 
 ****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

#include <png.h>


//------------------------

typedef struct
{
	int sleep_time,
		flags;
}OptionData;

typedef struct
{
	Display *xdisp;

	FILE *fp;
	unsigned char *rowbuf;
	int imgw,imgh;

	png_struct *png;
	png_info *pnginfo;
}WorkData;


static OptionData g_opt;
static WorkData g_work;

//------------------------

#define OPTF_ROOT_WINDOW  1
#define OPTF_NO_FRAME     2

//------------------------

#define GET_ATOM(name)  XInternAtom(g_work.xdisp, name, False)

//------------------------



//=============================
// sub
//=============================


/** エラー終了 */

void _exit_err(const char *mes)
{
	fprintf(stderr, "error: %s\n", mes);

	if(g_work.png)
		png_destroy_write_struct(&g_work.png, (g_work.pnginfo)? &g_work.pnginfo: 0);

	if(g_work.fp && g_work.fp != stdout)
		fclose(g_work.fp);

	if(g_work.rowbuf)
		free(g_work.rowbuf);
	
	if(g_work.xdisp)
		XCloseDisplay(g_work.xdisp);
	
	exit(1);
}


//=============================
// PNG
//=============================


/** PNG 出力初期化 (8bit RGB) */

void _png_output_init(const char *filename)
{
	png_struct *png;
	png_info *pnginfo;

	//ファイル開く

	if(strcmp(filename, "-") == 0)
		g_work.fp = stdout;
	else
	{
		g_work.fp = fopen(filename, "wb");
		if(!g_work.fp)
			_exit_err("cannot open output file");
	}
	
	//PNG 初期化
	
	g_work.png = png = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if(!png)
		_exit_err("libpng create");
	
	g_work.pnginfo = pnginfo = png_create_info_struct(png);
	if(!pnginfo)
		_exit_err("libpng create info");
	
	if(setjmp(png_jmpbuf(png)))
		_exit_err("libpng");
	
	png_init_io(png, g_work.fp);
	
	//IHDR
	
	png_set_IHDR(png, pnginfo, g_work.imgw, g_work.imgh,
		8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	png_write_info(png, pnginfo);

	//Y1行バッファ確保

	g_work.rowbuf = (unsigned char *)malloc(g_work.imgw * 3);
	if(!g_work.rowbuf)
		_exit_err("not enough memory");
}

/** PNG 出力終了 */

void _png_output_finish(void)
{
	png_write_end(g_work.png, g_work.pnginfo);
	png_destroy_write_struct(&g_work.png, &g_work.pnginfo);

	free(g_work.rowbuf);

	fclose(g_work.fp);
}


//=============================
// イメージ変換して出力
//=============================


/** RGB のビット位置取得 */

unsigned char _get_rgb_bitpos(unsigned int mask)
{
	unsigned int i;
	
	for(i = 0; i < 32; i++)
	{
		if(mask & (1 << i)) break;
	}
	
	return i;
}

/** RGB 8bit 以下の場合の左シフト数取得 */

unsigned char _get_rgb_lshift(unsigned int mask)
{
	unsigned int i;
	
	for(i = 0; i < 8; i++)
	{
		if(!(mask & (1 << i))) break;
	}
	
	return 8 - i;
}

/** XImage から PNG イメージ出力 */

void _output_image(XImage *ximg)
{
	int ix,iy,bpp,r,g,b,maxr,maxg,maxb,prog_cur,prog;
	unsigned char *ps,*pd,*psY;
	unsigned char bigendian,shift_r,shift_g,shift_b,lshift_r,lshift_g,lshift_b;
	unsigned int col;

	//情報
	
	bpp = ximg->bits_per_pixel >> 3;
	bigendian = (ximg->byte_order == MSBFirst);
	
	shift_r = _get_rgb_bitpos(ximg->red_mask);
	shift_g = _get_rgb_bitpos(ximg->green_mask);
	shift_b = _get_rgb_bitpos(ximg->blue_mask);
	
	lshift_r = _get_rgb_lshift(ximg->red_mask >> shift_r);
	lshift_g = _get_rgb_lshift(ximg->green_mask >> shift_g);
	lshift_b = _get_rgb_lshift(ximg->blue_mask >> shift_b);

	maxr = (1 << (8 - lshift_r)) - 1;
	maxg = (1 << (8 - lshift_g)) - 1;
	maxb = (1 << (8 - lshift_b)) - 1;
	
	//変換
	
	psY = (unsigned char *)ximg->data;
	prog_cur = 0;
	
	for(iy = 0; iy < g_work.imgh; iy++)
	{
		ps = psY;
		pd = g_work.rowbuf;
	
		for(ix = g_work.imgw; ix > 0; ix--)
		{
			//色値取得
			
			switch(bpp)
			{
				//32bit
				case 4:
					if(bigendian)
						col = ((unsigned int)ps[0] << 24) | (ps[1] << 16) | (ps[2] << 8) | ps[3];
					else
						col = ((unsigned int)ps[3] << 24) | (ps[2] << 16) | (ps[1] << 8) | ps[0];
					break;
				//24bit
				case 3:
					if(bigendian)
						col = ((unsigned int)ps[0] << 16) | (ps[1] << 8) | ps[2];
					else
						col = ((unsigned int)ps[2] << 16) | (ps[1] << 8) | ps[0];
					break;
				//16bit
				case 2:
					if(bigendian)
						col = ((unsigned int)ps[0] << 8) | ps[1];
					else
						col = ((unsigned int)ps[1] << 8) | ps[0];
					break;
			}

			//RGB 8bit
			
			r = (col & ximg->red_mask) >> shift_r;
			g = (col & ximg->green_mask) >> shift_g;
			b = (col & ximg->blue_mask) >> shift_b;

			if(bpp == 2)
			{
				if(r == maxr) r = 255;
				else r = r << lshift_r;

				if(g == maxg) g = 255;
				else g = g << lshift_g;

				if(b == maxb) b = 255;
				else b = b << lshift_b;
			}

			//
			
			pd[0] = r;
			pd[1] = g;
			pd[2] = b;
			
			ps += bpp;
			pd += 3;
		}

		//PNG にY1行出力

		png_write_row(g_work.png, g_work.rowbuf);
		
		psY += ximg->bytes_per_line;

		//進捗

		prog = iy * 5 / (g_work.imgh - 1);

		if(prog != prog_cur)
		{
			for(; prog_cur < prog; prog_cur++)
				putc('.', stderr);

			fflush(stderr);
		}
	}
}


//=============================
// X 関連
//=============================


/** カレントウィンドウ取得
 *
 * return: 0 で失敗 */

Window _get_current_window(Window root)
{
	Atom type;
	int format;
	unsigned long num,after;
	unsigned char *pdat;
	Window win = 0;

	if(XGetWindowProperty(g_work.xdisp, root,
		GET_ATOM("_NET_ACTIVE_WINDOW"), 0, 1024, False,
		XA_WINDOW, &type, &format, &num, &after, &pdat) == Success)
	{
		if(type == XA_WINDOW && format == 32)
			win = *((Atom *)pdat);
	
		XFree(pdat);
	}
	
	return win;
}

/** フレームウィンドウ取得 */

Window _get_frame_window(Window id)
{
	Window root,parent,*child;
	unsigned int num;

	//親 = ルートになるまで親をたどる

	while(1)
	{
		XQueryTree(g_work.xdisp, id, &root, &parent, &child, &num);
		if(child) XFree(child);

		if(parent == root) break;

		id = parent;
	}

	return id;
}

/** ウィンドウの位置とサイズを取得 */

void _get_window_info(Window win,int *dst)
{
	Window root,child;
	int x,y;
	unsigned int w,h,bw,depth;

	XGetGeometry(g_work.xdisp, win, &root, &x, &y, &w, &h, &bw, &depth);
	
	XTranslateCoordinates(g_work.xdisp, win, root, 0, 0, &x, &y, &child);
	
	dst[0] = x;
	dst[1] = y;
	dst[2] = w;
	dst[3] = h;
}

/** イメージ取得 & 出力 */

void _output(const char *filename)
{
	Display *disp;
	Window root,win;
	XImage *ximg;
	int depth,val[4];
	
	//ディスプレイ開く

	disp = XOpenDisplay(0);
	if(!disp)
		_exit_err("failed XOpenDisplay()");

	g_work.xdisp = disp;
	
	root = XDefaultRootWindow(disp);
	depth = DefaultDepth(disp, DefaultScreen(disp));
	
	if(depth <= 8)
		_exit_err("low display depth");
	
	//スリープ
	
	if(g_opt.sleep_time)
	{
		fprintf(stderr, "sleep %ds\n", g_opt.sleep_time);
		fflush(stderr);
		
		sleep(g_opt.sleep_time);
	}
	
	//対象ウィンドウ取得
	
	if(g_opt.flags & OPTF_ROOT_WINDOW)
		win = root;
	else
	{
		win = _get_current_window(root);
		if(win == 0)
			_exit_err("cannot get current window");

		//フレーム有りの場合はフレームウィンドウ取得

		if(!(g_opt.flags & OPTF_NO_FRAME))
			win = _get_frame_window(win);
	}
	
	//ウィンドウの位置とサイズ取得
	
	_get_window_info(win, val);

	//情報
	
	fprintf(stderr, "0x%lx (%d,%d)-[%dx%d] %dbit\n",
		win, val[0], val[1], val[2], val[3], depth);
	
	//XImage 取得
	
	ximg = XGetImage(disp, root, val[0], val[1], val[2], val[3], -1L, ZPixmap);
	if(!ximg)
		_exit_err("failed XGetImage()");

	g_work.imgw = ximg->width;
	g_work.imgh = ximg->height;

	//PNG 出力

	_png_output_init(filename);

	fprintf(stderr, "save");

	_output_image(ximg);

	_png_output_finish();

	fprintf(stderr, "OK\n");
		
	//終了

	XDestroyImage(ximg);

	XCloseDisplay(disp);
}


//=============================
// main
//=============================


/** ヘルプ表示 */

void _put_help(char *exename)
{
	fprintf(stderr,
"Usage: %s [opt] <output filename or '-'>\n\n"
"Options:\n"
"  -r       root window\n"
"  -f       no window frame\n"
"  -s [sec] sleep time (default=2)\n"
"  -h       show this help message\n"
	, exename);
	
	exit(0);
}

/** オプション取得 */

void _get_options(int argc,char **argv)
{
	int opt;

	while(1)
	{
		opt = getopt(argc, argv, "rfhs:");
		if(opt < 0) break;
		
		switch(opt)
		{
			case 'r':
				g_opt.flags |= OPTF_ROOT_WINDOW;
				break;
			case 'f':
				g_opt.flags |= OPTF_NO_FRAME;
				break;
			case 's':
				g_opt.sleep_time = atoi(optarg);
				break;
			case 'h':
				_put_help(argv[0]);
				break;
		}
	}
	
	if(optind >= argc)
		_put_help(argv[0]);
}

/** メイン */

int main(int argc,char **argv)
{
	//データ初期化

	g_opt.sleep_time = 2;
	g_opt.flags = 0;

	memset(&g_work, 0, sizeof(WorkData));

	//オプション取得

	_get_options(argc, argv);

	//出力
	
	_output(argv[optind]);

	return 0;
}
